# frozen_string_literal: true

#-------------------------------------------------------------------------
# # Copyright (c) Microsoft and contributors. All rights reserved.
#
# The MIT License(MIT)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions :

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#--------------------------------------------------------------------------

require "rubygems"
require "nokogiri"
require "base64"

require "azure/storage/common"

module Azure
  module Storage
    module Blob
      autoload :Default,       "azure/storage/blob/default"
      autoload :Version,       "azure/storage/blob/version"
      autoload :Blob,          "azure/storage/blob/blob"
      autoload :Block,         "azure/storage/blob/block"
      autoload :Page,          "azure/storage/blob/page"
      autoload :Append,        "azure/storage/blob/append"
      autoload :Container,     "azure/storage/blob/container"
      autoload :Serialization, "azure/storage/blob/serialization"
      autoload :BlobService,   "azure/storage/blob/blob_service"
    end
  end
end
